/**
 * File server plugin 
 *
 * By 
 *
 * Released under the GPL.
 *
 */
#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <libmpd/debug_printf.h>
#include <gmpc/plugin.h>
#include <gmpc/gmpc-profiles.h>
#include <libmpd/libmpd-internal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "microhttpd/microhttpd.h"
#include <config.h>

#define URLS_CLASS "Music"

gmpcPlugin plugin;



static struct MHD_Daemon * d = NULL;

static GtkListStore *ls = NULL;
static GtkWidget *mserver_vbox = NULL;
static config_obj *cfg_urls = NULL;

static GtkTreeRowReference *mserver_ref = NULL;
void mserver_browser_activated(GtkWidget *tree,GtkTreePath *path);

void mserver_browser_add(GtkWidget *cat_tree)
{
    GtkTreePath *path;
    GtkTreeIter iter;
    gint pos = cfg_get_single_value_as_int_with_default(config, "mserver","position",20);
    GtkTreeStore *pl3_tree = (GtkTreeStore  *)gtk_tree_view_get_model(GTK_TREE_VIEW(cat_tree));

    playlist3_insert_browser(&iter, pos);
	gtk_tree_store_set(pl3_tree, &iter, 
			PL3_CAT_TYPE, plugin.id,
			PL3_CAT_TITLE, "Serve music",
			PL3_CAT_INT_ID, "",
			PL3_CAT_ICON_ID, "gtk-open",
			PL3_CAT_PROC, TRUE,
			PL3_CAT_ICON_SIZE,GTK_ICON_SIZE_DND,-1);
    /**
     * Clean up old row reference if it exists
     */
    if (mserver_ref)
    {
        gtk_tree_row_reference_free(mserver_ref);
        mserver_ref = NULL;
    }
    /**
     * create row reference
     */
    path = gtk_tree_model_get_path(GTK_TREE_MODEL(playlist3_get_category_tree_store()), &iter);
    if (path)
    {
        mserver_ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(playlist3_get_category_tree_store()), path);
        gtk_tree_path_free(path);
    }

}

static void mserver_save_myself(void)
{
	if (mserver_ref)
    {
        GtkTreePath *path = gtk_tree_row_reference_get_path(mserver_ref);
        if(path)
        {
            gint *indices = gtk_tree_path_get_indices(path);
            debug_printf(DEBUG_INFO,"Saving myself to position: %i\n", indices[0]);
            cfg_set_single_value_as_int(config, "mserver","position",indices[0]);
            gtk_tree_path_free(path);
        }
    }
}



/**
 * Get/Set enabled
 */
int mserver_get_enabled()
{
	return cfg_get_single_value_as_int_with_default(config, "mserver", "enable", TRUE);
}
void mserver_set_enabled(int enabled)
{
	cfg_set_single_value_as_int(config, "mserver", "enable", enabled);
	if (enabled)
	{
		if(mserver_ref == NULL)
		{
			mserver_browser_add(GTK_WIDGET(playlist3_get_category_tree_view()));
		}
	}
	else if (mserver_ref)
	{
		GtkTreePath *path = gtk_tree_row_reference_get_path(mserver_ref);
		if (path){
			GtkTreeIter iter;

            mserver_save_myself();
			if (gtk_tree_model_get_iter(GTK_TREE_MODEL(playlist3_get_category_tree_store()), &iter, path)){
				gtk_tree_store_remove(playlist3_get_category_tree_store(), &iter);
			}
			gtk_tree_path_free(path);
			gtk_tree_row_reference_free(mserver_ref);
			mserver_ref = NULL;
		}                                                                                                  	
	}                                                                                                      	
	pl3_update_go_menu();
}

typedef struct {
	FILE *fp;
	size_t size;
}str_block;

static int file_reader(str_block *file, size_t pos, char * buf, int max) 
{
	int retv;
	retv = fread(buf,sizeof(char), max, file->fp);
	if(feof(file->fp))
	{
		return -1;
	}
	if(ferror(file->fp))
	{
		return -1;
	}
	/* hack to tell end of file, feof seems to fail..*/
	/* somehow this fixes stuff for mpd, but breaks it for f.e. wget */
	if((pos+retv) >= file->size)
	{
		return -1;
	}
	return retv;
}
static int apc_all(void * cls,
		const struct sockaddr * addr,
		socklen_t addrlen) {

	return MHD_YES; /* accept connections from anyone */
}
static void file_close(str_block *str)
{
	printf("closing stream\n");
	fclose(str->fp);
	g_free(str);
}
static int ahc_echo(void *nothing,
		struct MHD_Connection * connection,
		const char * url,
		const char * method,
        const char *version,
		const char * upload_data,
		unsigned int * upload_data_size,
        void **con_cls) {
	struct MHD_Response * response;
	int ret;

	if (0 != strcmp(method, "GET"))
		return MHD_NO; /* unexpected method */
	printf("url: '%s'\n", url);
	gchar *me = cfg_get_single_value_as_string(cfg_urls, URLS_CLASS, (char *)&url[1]);
	if(me && g_file_test(me, G_FILE_TEST_EXISTS))
	{
		/* needs error checking */
		gchar *mime_type = "application/octet-stream";
		gchar  *extention;
		const char *position;
		int pos;
		struct stat stats;
		str_block *str = g_malloc0(sizeof(str_block));
		stat(me, &stats);

		str->fp=fopen(me, "r");
		str->size = stats.st_size;
		response = MHD_create_response_from_callback(stats.st_size,
				4048,
				(MHD_ContentReaderCallback)&file_reader, 
				str,
				(MHD_ContentReaderFreeCallback)&file_close);
		position = MHD_lookup_connection_value (connection, MHD_HEADER_KIND, MHD_HTTP_HEADER_RANGE);
		if(position)
		{
			pos = atoi(&position[6]);
			fseek(str->fp,pos, SEEK_SET);
		}
		/* somehow gcc thinks I am not using this value, stupid gcc */
		for(extention = &me[strlen(me)]; me != extention && extention[0] != '.';*extention--);
		if(0 == strncasecmp(extention, ".flac",5))
		{
			mime_type = "audio/x-flac";
		}else if (0 == strncasecmp(extention, ".mp3", 4))
		{
			mime_type = "audio/mpeg";
		}else if (0 == strncasecmp(extention, ".ogg", 4))
		{
			mime_type = "audio/x-vorbis+ogg";
		} else if (0 == strncasecmp(extention, ".wv", 3))
		{
			mime_type = "audio/x-wavpack";
		}

		MHD_add_response_header(response,MHD_HTTP_HEADER_CONTENT_TYPE,mime_type); 
		extention = g_path_get_basename(me);
		MHD_add_response_header(response,"x-audiocast-name" , extention);
		g_free(extention);
		ret = MHD_queue_response(connection,
				MHD_HTTP_OK,
				response);
		MHD_destroy_response(response);
		g_free(me);
		return ret;

	}
	if(me)g_free(me);
	return MHD_NO;
}
void mserver_destroy()
{
	if(d)
	{
		MHD_stop_daemon(d);
		d = NULL;
	}
	if(ls)
	{
		gtk_list_store_clear(ls);
		g_object_unref(ls);
		ls = NULL;
	}
	if(mserver_vbox)
	{
		gtk_widget_destroy(mserver_vbox);
	}
	if(cfg_urls)
	{
		cfg_close(cfg_urls);
		cfg_urls = NULL;
	}
}
void mserver_browser_add_file()
{
	GtkWidget *dialog = gtk_file_chooser_dialog_new("Add File", NULL, GTK_FILE_CHOOSER_ACTION_OPEN,GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OK, GTK_RESPONSE_OK,NULL);

	GtkFileFilter *filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, "all");
	gtk_file_filter_add_pattern(filter,"*.ogg");
	gtk_file_filter_add_pattern(filter,"*.mp3");
	gtk_file_filter_add_pattern(filter,"*.flac");
	gtk_file_filter_add_pattern(filter,"*.wv");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, "ogg");
	gtk_file_filter_add_pattern(filter,"*.ogg");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);
	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, "mp3");
	gtk_file_filter_add_pattern(filter,"*.mp3");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);
	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, "flac");
	gtk_file_filter_add_pattern(filter,"*.flac");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);
	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, "wavpack");
	gtk_file_filter_add_pattern(filter,"*.wv");                 	
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);

	gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(dialog),TRUE);



	gtk_widget_show(dialog);
	switch(gtk_dialog_run(GTK_DIALOG(dialog)))
	{
		case GTK_RESPONSE_OK:
			{
				GSList *filenames = gtk_file_chooser_get_filenames(GTK_FILE_CHOOSER(dialog));
				if(filenames)
				{
					GSList *iter = filenames; 
					for(;iter;iter= g_slist_next(iter))
					{
						GtkTreeIter titer;
						gchar *filename = iter->data;
						guint id = g_random_int();
						gchar *name = g_strdup_printf("%u", id);
						gtk_list_store_append(GTK_LIST_STORE(ls), &titer);
						gtk_list_store_set(ls, &titer, 0,id, 1, filename, -1);

						cfg_set_single_value_as_string(cfg_urls,URLS_CLASS, name, filename);			
						g_free(name);
					}
					g_slist_foreach(filenames, (GFunc)g_free, NULL);
					g_slist_free(filenames);
				}
			}
		default:
			break;
	}
	gtk_widget_destroy(dialog);
}
void mserver_browser_activated(GtkWidget *tree, GtkTreePath *path)
{
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
	GtkTreeIter iter;
	if(gtk_tree_model_get_iter(model, &iter,path))
	{		
#ifdef HAVE_SOCKADDR_IN6_STRUCT
		struct sockaddr_in6 name;
#else /* HAVE_SOCKADDR_IN6_STRUCT */
		struct sockaddr_in name;
#endif /* HAVE_SOCKADDR_IN6_STRUCT */


		socklen_t namelen = sizeof(name);
		gchar *peername;
		if(getsockname(connection->connection->sock,(struct sockaddr*)&name, &namelen)<0){
			peername = g_strdup("localhost");
		} else{
			peername = g_strdup(inet_ntoa(name.sin_addr));
		}

		guint uid;
		gchar *url = NULL;
		gtk_tree_model_get(GTK_TREE_MODEL(model), &iter, 0, &uid, -1);
		url = g_strdup_printf("http://%s:9876/%u",peername, uid);
		g_free(peername);
		mpd_playlist_add(connection, url);
		g_free(url);
	}
}
static void mserver_browser_remove_files(GtkButton *button, GtkTreeView *tree)
{
	GtkTreeModel *model = gtk_tree_view_get_model(tree);
	GtkTreeSelection *sel = gtk_tree_view_get_selection(tree);
	GList *del_items = NULL;
	GList *iter, *list = gtk_tree_selection_get_selected_rows(sel, &model);
	for(iter = list; iter;iter = g_list_next(iter))
	{
		GtkTreeIter titer;
		if(gtk_tree_model_get_iter(model, &titer, iter->data))
		{
			guint uid;
			gchar *uids;
			gtk_tree_model_get(GTK_TREE_MODEL(model), &titer, 0, &uid, -1);
			uids = g_strdup_printf("%u", uid);
			cfg_del_single_value(cfg_urls, URLS_CLASS, uids);
			g_free(uids);
			del_items = g_list_append(del_items, gtk_tree_row_reference_new(model, iter->data));
		}
	}
	g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
	g_list_free(list);

	iter = g_list_first(del_items);
	for(;iter;iter= g_list_next(iter))
	{
		GtkTreeIter titer;
		if(gtk_tree_model_get_iter(model, &titer, gtk_tree_row_reference_get_path(iter->data)))
		{
			gtk_list_store_remove(GTK_LIST_STORE(model), &titer);	
		}
		gtk_tree_row_reference_free(iter->data);
	}
	g_list_free(del_items);
}

static void mserver_browser_add_files_to_playlist(GtkButton *button, GtkTreeView *tree)
{
	GtkTreeModel *model = gtk_tree_view_get_model(tree);
	GtkTreeSelection *sel = gtk_tree_view_get_selection(tree);
	GList *iter, *list = gtk_tree_selection_get_selected_rows(sel, &model);
	for(iter = list; iter;iter = g_list_next(iter))
	{
		GtkTreeIter titer;
		if(gtk_tree_model_get_iter(model, &titer, iter->data))
		{
#ifdef HAVE_SOCKADDR_IN6_STRUCT
			struct sockaddr_in6 name;
#else /* HAVE_SOCKADDR_IN6_STRUCT */
			struct sockaddr_in name;
#endif /* HAVE_SOCKADDR_IN6_STRUCT */
			guint uid;
			gchar *url = NULL;
			socklen_t namelen = sizeof(name);
			gchar *peername;
			if(getsockname(connection->connection->sock,(struct sockaddr*)&name, &namelen)<0){
				peername = g_strdup("localhost");
			} else{
				peername = g_strdup(inet_ntoa(name.sin_addr));
			}
			gtk_tree_model_get(GTK_TREE_MODEL(model), &titer, 0, &uid, -1);
			url = g_strdup_printf("http://%s:9876/%u",peername, uid);
			g_free(peername);
			mpd_playlist_add(connection, url);
			g_free(url);
		}
	}
	g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
	g_list_free(list);
}

void mserver_init() {
	GtkWidget *sw, *tree,*button,*bbox;
	GtkCellRenderer *renderer;
	conf_mult_obj *cmo,*cmo_iter	;
	GtkTreeIter iter;
	gchar *fs;
	/* config file */
	fs = gmpc_get_user_path("server_urls.txt");
	cfg_urls = cfg_open(fs);
	g_free(fs);	

	printf("Start deamon\n");
	d = MHD_start_daemon(MHD_USE_THREAD_PER_CONNECTION,
			9876,	
			&apc_all,
			NULL,
			&ahc_echo,
			NULL,
			MHD_OPTION_END);

	ls = gtk_list_store_new(2, G_TYPE_UINT, G_TYPE_STRING);
	cmo= cmo_iter = cfg_get_key_list(cfg_urls, URLS_CLASS);
	if(cmo)
	{
		for(;cmo_iter;cmo_iter = cmo_iter->next)
		{
			guint id = atoi(cmo_iter->key);
			gtk_list_store_append(ls, &iter);
			gtk_list_store_set(ls, &iter, 0, id, 1, cmo_iter->value,-1);
		}
		cfg_free_multiple(cmo);
	}
	mserver_vbox = gtk_vbox_new(FALSE,6);
	/* Scrolled window */
	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_ETCHED_IN);
	/* TreeView */
	tree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(ls));
	gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(tree)), GTK_SELECTION_MULTIPLE);

	g_signal_connect(G_OBJECT(tree), "row-activated", G_CALLBACK(mserver_browser_activated), NULL);
	gtk_container_add(GTK_CONTAINER(sw),tree);
	gtk_box_pack_start(GTK_BOX(mserver_vbox), sw, TRUE, TRUE, 0); 

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,"Path",renderer,"text",1,NULL);

	bbox = gtk_hbutton_box_new();
	
	button = gtk_button_new_with_label("Add files");
	gtk_button_set_image(GTK_BUTTON(button), gtk_image_new_from_stock(GTK_STOCK_ADD, GTK_ICON_SIZE_BUTTON));
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(mserver_browser_add_file), NULL);
	gtk_box_pack_start(GTK_BOX(bbox), button,FALSE,FALSE,0);

	button = gtk_button_new_with_label("Add to playlist");
	gtk_button_set_image(GTK_BUTTON(button), gtk_image_new_from_stock(GTK_STOCK_ADD, GTK_ICON_SIZE_BUTTON));
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(mserver_browser_add_files_to_playlist), tree);
	gtk_box_pack_start(GTK_BOX(bbox), button,FALSE,FALSE,0);

	button = gtk_button_new_from_stock(GTK_STOCK_REMOVE);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(mserver_browser_remove_files),tree);
	gtk_box_pack_start(GTK_BOX(bbox), button,FALSE,FALSE,0);
	gtk_box_pack_start(GTK_BOX(mserver_vbox), bbox,FALSE,FALSE,0);


	g_object_ref(mserver_vbox);
	gtk_widget_show_all(mserver_vbox);
}


void mserver_browser_selected(GtkWidget *container)
{
	gtk_container_add(GTK_CONTAINER(container), mserver_vbox);
}

void mserver_browser_unselected(GtkWidget *container)
{
	gtk_container_remove(GTK_CONTAINER(container), mserver_vbox);
}

gmpcPlBrowserPlugin mserver_browser_plugin ={
	.add = mserver_browser_add,
	.selected = mserver_browser_selected,
	.unselected = mserver_browser_unselected,

};

int plugin_api_version = PLUGIN_API_VERSION;
gmpcPlugin plugin = {
	.name = "File Serve plugin",
	.version        = {PLUGIN_MAJOR_VERSION,PLUGIN_MINOR_VERSION,PLUGIN_MICRO_VERSION},
	.plugin_type = GMPC_PLUGIN_PL_BROWSER,
	.init = mserver_init,
	.destroy = mserver_destroy,
    .save_yourself = mserver_save_myself,
	.get_enabled = mserver_get_enabled,
	.set_enabled = mserver_set_enabled,
	.browser = &mserver_browser_plugin,
};
